package jalview.datamodel.ontology;

import jalview.io.gff.SequenceOntologyFactory;
import jalview.io.gff.SequenceOntologyI;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * A base class for models of Sequence Ontology and others
 * 
 * @author gmcarstairs
 *
 */
public abstract class OntologyBase implements OntologyI
{
  protected Map<String, List<String>> rootParents = new HashMap<>();

  @Override
  public Set<String> getParentTerms(Set<String> terms)
  {
    Set<String> parents = new HashSet<>(terms);

    boolean childRemoved = true;
    while (childRemoved)
    {
      childRemoved = removeChild(parents);
    }
    return parents;
  }

  /**
   * Removes the first term in the given set found which is a child of another
   * term in the set. Answers true if a child was found and removed, else false.
   * 
   * @param terms
   * @return
   */
  boolean removeChild(Set<String> terms)
  {
    for (String t1 : terms)
    {
      for (String t2 : terms)
      {
        if (t1 != t2)
        {
          if (isA(t1, t2))
          {
            terms.remove(t1);
            return true;
          }
          if (isA(t2, t1))
          {
            terms.remove(t2);
            return true;
          }
        }
      }
    }
    return false;
  }

  @Override
  public List<String> getChildTerms(String parent, List<String> terms)
  {
    List<String> children = new ArrayList<>();
    for (String term : terms)
    {
      if (!term.equals(parent) && isA(term, parent))
      {
        children.add(term);
      }
    }
    return children;
  }

  /**
   * Answers a (possibly empty) map of any Ontology terms (from the given term
   * and its parents) which subsume one or more of the target terms. The map key
   * is an ontology term, and the entry is the list of target terms that are
   * sub-terms of the key.
   * <p>
   * For example if {@code stop_gained} and {@code stop_lost} are known feature
   * types, then SO term {@ nonsynonymous_variant} is the first common parent of
   * both terms
   * 
   * @param givenTerm
   *          the term to search from
   * @param targetTerms
   *          candidate terms to 'capture' in ontology groupings
   * @return
   */
  @Override
  public Map<String, List<String>> findSequenceOntologyGroupings(
          String givenTerm, List<String> targetTerms)
  {
    List<String> sortedTypes = new ArrayList<>(targetTerms);
    Collections.sort(sortedTypes);
  
    Map<String, List<String>> parents = new HashMap<>();
    if (!isValidTerm(givenTerm))
    {
      return parents;
    }

    /*
     * method: 
     * walk up featureType and all of its parents
     * find other feature types which are subsumed by each term
     * add each distinct aggregation of included feature types to the map
     */
    List<String> candidates = new ArrayList<>();
    SequenceOntologyI so = SequenceOntologyFactory.getInstance();

    candidates.add(givenTerm);
    while (!candidates.isEmpty())
    {
      String term = candidates.remove(0);
      List<String> includedFeatures = new ArrayList<>();
      for (String type : sortedTypes)
      {
        if (!type.equals(givenTerm) && so.isA(type, term))
        {
          includedFeatures.add(type);
        }
      }
      if (!includedFeatures.isEmpty()
              && !parents.containsValue(includedFeatures))
      {
        parents.put(term, includedFeatures);
      }
      candidates.addAll(so.getParents(term));
    }
  
    return parents;
  }
}
