package jalview.datamodel.ontology;

import java.util.List;
import java.util.Map;
import java.util.Set;

public interface OntologyI
{
  /**
   * Answers true if the term can be identified in the ontology (possibly by id,
   * description or alias), else false
   * 
   * @param term
   * @return
   */
  boolean isValidTerm(String term);

  /**
   * Answers true if <code>childTerm</code> is the same as, or a sub-type
   * (specialisation of) <code>parentTerm</code>, else false
   * 
   * @param childTerm
   * @param parentTerm
   * @return
   */
  boolean isA(String childTerm, String parentTerm);

  /**
   * Answers those terms in the given set which are not child terms of some
   * other term in the set. That is, returns a set of parent terms. The input
   * set is not modified.
   * 
   * @param terms
   * @return
   */
  Set<String> getParentTerms(Set<String> terms);

  /**
   * Answers a (possibly empty) list of those terms in the supplied list which
   * are a child (directly or indirectly) of <code>parent</code>. The parent
   * term itself is not included (even if in the input list)
   * 
   * @param parent
   * @param terms
   * @return
   */
  List<String> getChildTerms(String parent, List<String> terms);

  /**
   * Answers a (possibly empty) list of the immediate parent terms of the given
   * term
   * 
   * @param term
   * @return
   */
  List<String> getParents(String term);

  /**
   * Returns a sorted list of all valid terms queried for (i.e. terms processed
   * which were valid in the SO), using the friendly description.
   * 
   * This can be used to check that any hard-coded stand-in for the full SO
   * includes all the terms needed for correct processing.
   * 
   * @return
   */
  List<String> termsFound();

  /**
   * Returns a sorted list of all invalid terms queried for (i.e. terms
   * processed which were not found in the SO), using the friendly description.
   * 
   * This can be used to report any 'non-compliance' in data, and/or to report
   * valid terms missing from any hard-coded stand-in for the full SO.
   * 
   * @return
   */
  List<String> termsNotFound();

  /**
   * Answers the top level parent terms (normally only one) for the given term,
   * that is, those that have no parent themselves. Answers null if {@code term}
   * is not a sequence ontology term. Answers a list just containing
   * {@code term} if it is a valid term with no parent.
   * 
   * @param term
   * @return
   */
  List<String> getRootParents(String term);

  /**
   * Answers a (possibly empty) map of any Ontology terms (from the given term
   * and its parents) which subsume one or more of the target terms. The map key
   * is an ontology term, and the entry is the list of target terms that are
   * sub-terms of the key.
   * <p>
   * For example if {@code stop_gained} and {@code stop_lost} are known feature
   * types, then SO term {@ nonsynonymous_variant} is the first common parent of
   * both terms
   * 
   * @param givenTerm
   *          the term to search from
   * @param targetTerms
   *          candidate terms to 'capture' in ontology groupings
   * @return
   */
  Map<String, List<String>> findSequenceOntologyGroupings(String givenTerm,
          List<String> targetTerms);
}