/*
 * Jalview - A Sequence Alignment Editor and Viewer ($$Version-Rel$$)
 * Copyright (C) $$Year-Rel$$ The Jalview Authors
 * 
 * This file is part of Jalview.
 * 
 * Jalview is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *  
 * Jalview is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty 
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR 
 * PURPOSE.  See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Jalview.  If not, see <http://www.gnu.org/licenses/>.
 * The Jalview Authors are detailed in the 'AUTHORS' file.
 */
package jalview.ext.so;

import jalview.bin.Cache;
import jalview.datamodel.ontology.OntologyBase;
import jalview.io.gff.SequenceOntologyI;

import java.io.BufferedInputStream;
import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.text.ParseException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.NoSuchElementException;
import java.util.Set;
import java.util.zip.ZipEntry;
import java.util.zip.ZipInputStream;

import org.biojava.nbio.ontology.Ontology;
import org.biojava.nbio.ontology.Synonym;
import org.biojava.nbio.ontology.Term;
import org.biojava.nbio.ontology.Term.Impl;
import org.biojava.nbio.ontology.Triple;
import org.biojava.nbio.ontology.io.OboParser;
import org.biojava.nbio.ontology.utils.Annotation;

/**
 * A wrapper class that parses the Sequence Ontology and exposes useful access
 * methods. This version uses the BioJava parser.
 */
public class SequenceOntology extends OntologyBase
        implements SequenceOntologyI
{
  /*
   * the parsed Ontology data as modelled by BioJava
   */
  private Ontology ontology;

  /*
   * the ontology term for the isA relationship
   */
  private Term isA;

  /*
   * lookup of terms by user readable name (NB not guaranteed unique)
   */
  private Map<String, Term> aliases;

  /*
   * Map where key is a Term and value is a (possibly empty) list of 
   * all Terms to which the key has an 'isA' relationship, either
   * directly or indirectly (A isA B isA C)
   */
  private Map<Term, List<Term>> termIsA;

  private List<String> termsFound;

  private List<String> termsNotFound;

  /**
   * Package private constructor to enforce use of singleton. Parses and caches
   * the SO OBO data file.
   */
  public SequenceOntology()
  {
    termsFound = new ArrayList<>();
    termsNotFound = new ArrayList<>();
    aliases = new HashMap<>();
    termIsA = new HashMap<>();

    loadOntologyZipFile("so-simple.obo");
  }

  /**
   * Loads the given ontology file from a zip file with ".zip" appended
   * 
   * @param ontologyFile
   */
  protected void loadOntologyZipFile(String ontologyFile)
  {
    long now = System.currentTimeMillis();
    ZipInputStream zipStream = null;
    try
    {
      String zipFile = ontologyFile + ".zip";
      InputStream inStream = this.getClass()
              .getResourceAsStream("/" + zipFile);
      zipStream = new ZipInputStream(new BufferedInputStream(inStream));
      ZipEntry entry;
      while ((entry = zipStream.getNextEntry()) != null)
      {
        if (entry.getName().equals(ontologyFile))
        {
          loadOboFile(zipStream);
        }
      }
      long elapsed = System.currentTimeMillis() - now;
      System.out.println("Loaded Sequence Ontology from " + zipFile + " ("
              + elapsed + "ms)");
    } catch (Exception e)
    {
      e.printStackTrace();
    } finally
    {
      closeStream(zipStream);
    }
  }

  /**
   * Closes the input stream, swallowing all exceptions
   * 
   * @param is
   */
  protected void closeStream(InputStream is)
  {
    if (is != null)
    {
      try
      {
        is.close();
      } catch (IOException e)
      {
        // ignore
      }
    }
  }

  /**
   * Reads, parses and stores the OBO file data
   * 
   * @param is
   * @throws ParseException
   * @throws IOException
   */
  protected void loadOboFile(InputStream is)
          throws ParseException, IOException
  {
    BufferedReader oboFile = new BufferedReader(new InputStreamReader(is));
    OboParser parser = new OboParser();
    ontology = parser.parseOBO(oboFile, "SO", "the SO ontology");
    isA = ontology.getTerm("is_a");
    storeTermAliases();
  }

  /**
   * Stores a lookup table of terms by description or synonym. Note that
   * description is not guaranteed unique. Where duplicate descriptions are
   * found, try to discard the term that is flagged as obsolete. However we do
   * store obsolete terms where there is no duplication of description.
   */
  protected void storeTermAliases()
  {
    Set<String> ambiguous = new HashSet<>();

    for (Term term : ontology.getTerms())
    {
      if (term instanceof Impl)
      {
        boolean newTermIsObsolete = isObsolete(term);
        String description = term.getDescription();
        if (description != null)
        {
          description = canonicalise(description);
          Term replaced = aliases.get(description);
          if (replaced != null)
          {
            boolean oldTermIsObsolete = isObsolete(replaced);
            if (newTermIsObsolete && !oldTermIsObsolete)
            {
              Cache.log.debug("SequenceOntology ignoring " + term.getName()
                      + " as obsolete and duplicated by "
                      + replaced.getName());
              term = replaced;
            }
            else if (!newTermIsObsolete && oldTermIsObsolete)
            {
              Cache.log.debug("SequenceOntology ignoring "
                      + replaced.getName()
                      + " as obsolete and duplicated by " + term.getName());
            }
            else
            {
              Cache.log.debug("SequenceOntology warning: " + term.getName()
                      + " has replaced " + replaced.getName()
                      + " for lookup of '" + description + "'");
            }
          }
          aliases.put(description, term);

          /*
           * also store synonyms if not ambiguous
           */
          if (!newTermIsObsolete)
          {
            storeSynonymsForTerm(term, ambiguous);
          }
        }
      }
    }

    /*
     * remove ambiguous synonyms for safety;
     * problem: what if a synonym matches a description?
     * only one case found:
     * nmd_transcript is synonym for SO:0001621:NMD_transcript_variant 
     * and also the description for SO:0002114:NMD_transcript
     */
    for (String syn : ambiguous)
    {
      aliases.remove(syn);
    }
  }

  /**
   * Stores any synonyms as an alternative lookup for the term, canonicalised
   * for case/hyphen/space insensitivity on lookup.
   * <p>
   * Some synonyms may be ambiguous (present for more than one term), and these
   * are handled as follows:
   * <ul>
   * <li>if a synonym matches the <em>description</em> of another term, it is
   * not saved, so that a term can always be found by description
   * <ul>
   * <li>Example: {@code nmd_transcript} is the description for
   * {@code NMD_transcript} and also a synonym for
   * {@code NMD_transcript_variant} - the synonym is ignored</li>
   * </ul>
   * </li>
   * <li>if one term is a sub-term (directly or indirectly) of the other, the
   * synonym is retained for the more general term
   * <ul>
   * <li>Example: {@code helix} is a synonym for
   * {@code alpha_helix, right_handed_peptide_helix, peptide_helix} - it is kept
   * for {@code peptide_helix} as this is a parent of the other terms</li>
   * </ul>
   * </li>
   * <li>otherwise the synonym is added to the {@code ambiguous} list for
   * removal
   * <ul>
   * <li>Example: {@code sequence variation} is a synonym for
   * {@code sequence_alteration} and {@code alternate_sequence_site} but these
   * have no {@code isA} relationship - the synonym is ignored as ambiguous</li>
   * </ul>
   * </ul>
   * 
   * @param term
   * @param ambiguous
   */
  void storeSynonymsForTerm(Term term, Set<String> ambiguous)
  {
    for (Object syn : term.getSynonyms())
    {
      String name = ((Synonym) syn).getName();
      String synonym = canonicalise(name);
      if (aliases.containsKey(synonym))
      {
        final Term found = aliases.get(synonym);
        if (found != term)
        {
          /*
           * this alias is ambiguous - matches description,
           * or an alias, of another term
           */
          String msg = String.format(
                  "SequenceOntology ambiguous synonym %s for '%s:%s' and '%s:%s'",
                  synonym, term.getName(), term.getDescription(),
                  found.getName(), found.getDescription());
          Cache.log.debug(msg);

          /*
           * preserve any entry whose canonical description happens to match
           * a synonym (NMD_transcript is a valid description, and also
           * a synonym for NMD_transcript_variant)
           * also preserve a parent (more general) term
           */
          if (synonym.equals(canonicalise(found.getDescription()))
                  || termIsA(term, found))
          {
            // leave it alone
          }
          /*
           * replace a specialised term with a more general one
           * with the same alias
           */
          // else if
          // (synonym.equals(canonicalise(term.getDescription())))
          else if (termIsA(found, term))
          {
            aliases.put(synonym, term);
          }
          else
          {
            ambiguous.add(synonym);
          }
        }
      }
      else
      {
        aliases.put(synonym, term);
      }
    }
  }

  /**
   * Converts a string to lower case and changes hyphens and spaces to
   * underscores
   * 
   * @param s
   * @return
   */
  static String canonicalise(String s)
  {
    return s == null ? null
            : s.toLowerCase().replace('-', '_').replace(' ', '_');
  }

  /**
   * Answers true if the term has property "is_obsolete" with value true, else
   * false
   * 
   * @param term
   * @return
   */
  public static boolean isObsolete(Term term)
  {
    Annotation ann = term.getAnnotation();
    if (ann != null)
    {
      try
      {
        if (Boolean.TRUE.equals(ann.getProperty("is_obsolete")))
        {
          return true;
        }
      } catch (NoSuchElementException e)
      {
        // fall through to false
      }
    }
    return false;
  }

  /**
   * Test whether the given Sequence Ontology term is nucleotide_match (either
   * directly or via is_a relationship)
   * 
   * @param soTerm
   *          SO name or description
   * @return
   */
  public boolean isNucleotideMatch(String soTerm)
  {
    return isA(soTerm, NUCLEOTIDE_MATCH);
  }

  /**
   * Test whether the given Sequence Ontology term is protein_match (either
   * directly or via is_a relationship)
   * 
   * @param soTerm
   *          SO name or description
   * @return
   */
  public boolean isProteinMatch(String soTerm)
  {
    return isA(soTerm, PROTEIN_MATCH);
  }

  /**
   * Test whether the given Sequence Ontology term is polypeptide (either
   * directly or via is_a relationship)
   * 
   * @param soTerm
   *          SO name or description
   * @return
   */
  public boolean isPolypeptide(String soTerm)
  {
    return isA(soTerm, POLYPEPTIDE);
  }

  /**
   * Returns true if the given term has a (direct or indirect) 'isA'
   * relationship with the parent
   * 
   * @param child
   * @param parent
   * @return
   */
  @Override
  public boolean isA(String child, String parent)
  {
    if (child == null || parent == null)
    {
      return false;
    }
    /*
     * optimise trivial checks like isA("CDS", "CDS")
     */
    if (child.equals(parent))
    {
      termFound(child);
      return true;
    }

    Term childTerm = getTerm(child);
    if (childTerm != null)
    {
      termFound(child);
    }
    else
    {
      termNotFound(child);
    }
    Term parentTerm = getTerm(parent);

    return termIsA(childTerm, parentTerm);
  }

  /**
   * Records a valid term queried for, for reporting purposes
   * 
   * @param term
   */
  private void termFound(String term)
  {
    synchronized (termsFound)
    {
      if (!termsFound.contains(term))
      {
        termsFound.add(term);
      }
    }
  }

  /**
   * Records an invalid term queried for, for reporting purposes
   * 
   * @param term
   */
  private void termNotFound(String term)
  {
    synchronized (termsNotFound)
    {
      if (!termsNotFound.contains(term))
      {
        Cache.log.debug("SequenceOntology term " + term + " invalid");
        termsNotFound.add(term);
      }
    }
  }

  /**
   * Returns true if the childTerm 'isA' parentTerm (directly or indirectly).
   * 
   * @param childTerm
   * @param parentTerm
   * @return
   */
  protected synchronized boolean termIsA(Term childTerm, Term parentTerm)
  {
    /*
     * null term could arise from a misspelled SO description
     */
    if (childTerm == null || parentTerm == null)
    {
      return false;
    }

    /*
     * recursive search endpoint:
     */
    if (childTerm == parentTerm)
    {
      return true;
    }

    /*
     * lazy initialisation - find all of a term's parents (recursively) 
     * the first time this is called, and save them in a map.
     */
    if (!termIsA.containsKey(childTerm))
    {
      findParents(childTerm);
    }

    List<Term> parents = termIsA.get(childTerm);
    for (Term parent : parents)
    {
      if (termIsA(parent, parentTerm))
      {
        /*
         * add (great-)grandparents to parents list as they are discovered,
         * for faster lookup next time
         */
        if (!parents.contains(parentTerm))
        {
          parents.add(parentTerm);
        }
        return true;
      }
    }

    return false;
  }

  /**
   * Finds all the 'isA' parents of the childTerm and stores them as a (possibly
   * empty) list.
   * 
   * @param childTerm
   */
  protected synchronized void findParents(Term childTerm)
  {
    List<Term> result = new ArrayList<>();
    for (Triple triple : ontology.getTriples(childTerm, null, isA))
    {
      Term parent = triple.getObject();
      result.add(parent);

      /*
       * and search for the parent's parents recursively
       */
      findParents(parent);
    }
    termIsA.put(childTerm, result);
  }

  /**
   * Returns the Term for a given name (e.g. "SO:0000735") or description (e.g.
   * "sequence_location"), or alias, or null if not found
   * 
   * @param child
   * @return
   */
  protected Term getTerm(final String nameOrDescription)
  {
    if (nameOrDescription == null)
    {
      return null;
    }
    Term t = aliases.get(canonicalise(nameOrDescription));
    if (t == null)
    {
      try
      {
        t = ontology.getTerm(nameOrDescription);
      } catch (NoSuchElementException e)
      {
        // not found
      }
    }
    return t;
  }

  public boolean isSequenceVariant(String soTerm)
  {
    return isA(soTerm, SEQUENCE_VARIANT);
  }

  /**
   * Sorts (case-insensitive) and returns the list of valid terms queried for
   */
  @Override
  public List<String> termsFound()
  {
    synchronized (termsFound)
    {
      Collections.sort(termsFound, String.CASE_INSENSITIVE_ORDER);
      return termsFound;
    }
  }

  /**
   * Sorts (case-insensitive) and returns the list of invalid terms queried for
   */
  @Override
  public List<String> termsNotFound()
  {
    synchronized (termsNotFound)
    {
      Collections.sort(termsNotFound, String.CASE_INSENSITIVE_ORDER);
      return termsNotFound;
    }
  }

  /**
   * {@inheritDoc}
   * 
   * @throws IllegalStateException
   *           if a loop is detected in the ontology
   */
  @Override
  public List<String> getRootParents(final String term)
  {
    /*
     * check in cache first
     */
    if (rootParents.containsKey(term))
    {
      return rootParents.get(term);
    }
    Term t = getTerm(term);
    if (t == null)
    {
      return null;
    }

    /*
     * todo: check for loops using 'seen', allowing for alternate paths e.g.
     * stop_gained isA feature_truncation isA feature_variant
     * " isA nonsynonymous_variant ... isA geneVariant isA feature_variant 
     */
    List<Term> seen = new ArrayList<>();
    List<Term> top = new ArrayList<>();
    List<Term> query = new ArrayList<>();
    query.add(t);

    while (!query.isEmpty())
    {
      List<Term> nextQuery = new ArrayList<>();
      for (Term q : query)
      {
        Set<Triple> parents = ontology.getTriples(q, null, isA);
        if (parents.isEmpty())
        {
          /*
           * q has no parents so is a top level term
           */
          top.add(q);
        }
        else
        {
          /*
           * search all parent terms
           */
          for (Triple triple : parents)
          {
            Term parent = triple.getObject();
            nextQuery.add(parent);
          }
        }
      }
      query = nextQuery;
    }

    List<String> result = new ArrayList<>();
    for (Term found : top)
    {
      String desc = found.getDescription();
      if (!result.contains(desc))
      {
        result.add(desc);
      }
    }

    /*
     * save result in cache
     */
    rootParents.put(term, result);

    return result;
  }

  @Override
  public List<String> getParents(String term)
  {
    List<String> parents = new ArrayList<>();
    Term t = getTerm(term);
    if (t != null)
    {
      for (Triple triple : ontology.getTriples(t, null, isA))
      {
        Term parent = triple.getObject();
        parents.add(parent.getDescription());
      }
    }
    return parents;
  }

  @Override
  public boolean isValidTerm(String term)
  {
    return getTerm(term) != null;
  }
}
